# Maintainer:     Ryan Young
# Last Modified:  May 06, 2022
from sodapy import Socrata
import requests

import pandas as pd

class Soda:
    """
    A basic wrapper for sodapy, which is a wrapper to communicate
    with Socrata API.
    ---
    Holds the user's app token as a class variable, and wraps
    important functions by declaring a new Socrata() object client
    with each method call.
    """

    token:str = None
    alt_token:str = None

    @classmethod
    def client_get(cls, base_url, data_id, **kwargs):
        """
        Try to make request. If request fails, try with alt token
        """
        try:
            client = Socrata(base_url, cls.token)
            return client.get(data_id, **kwargs)
        except Exception as e:
            client = Socrata(base_url, cls.alt_token)
            return client.get(data_id, **kwargs)


    @classmethod
    def client_get_metadata(cls, base_url, data_id, content_type):
        """
        Try to make request. If request fails, try with alt token
        """
        try:
            client = Socrata(base_url, cls.token)
            return client.get_metadata(data_id, content_type)
        except Exception as e:
            client = Socrata(base_url, cls.alt_token)
            return client.get_metadata(data_id, content_type)


    @classmethod
    def get(cls, base_url:str, data_id:str, full=False, suppress_warning=False, **kwargs) -> pd.DataFrame:
        """
        Creates a Socrata client using base url, and then
        retrieves data with data_id, passing any additional kwargs
        to client.get()
        ---
        Returns: pd.DataFrame
        """
        if 'limit' not in kwargs:
            if full == False:
                limit = 5
                if suppress_warning == False:
                    print("Pass 'full=True' to get full dataset.")
            else:
                limit = 10_000_000

        if 'limit' in kwargs:
            data = cls.client_get(base_url, data_id, **kwargs)
        else:
            data = cls.client_get(base_url, data_id, limit=limit, **kwargs)

        return pd.DataFrame.from_records(data)


    @classmethod
    def get_metadata(cls, base_url:str, data_id:str, content_type="json") -> dict:
        """
        Works just like get, except doesn't accept keyword args,
        and returns the dataset's metadata as a dictionary, instead
        of dataframe.
        """
        try:
            data = cls.client_get_metadata(base_url, data_id, content_type)
        except Exception as e:
            print("INVALID ID")
            return

        return dict(data)




# Maintainer:     Ryan Young
# Last Modified:  May 08, 2022
import requests
import pandas as pd
import numpy as np

class MetaLibrary(type):
    """
    This is my solution to effectively applying the property
    decorator to class variables. This allows us to execute
    logic when the Library's variables are accessed. In this
    case, the 'data' variable won't be loaded with online api
    data until the variable is actually accessed for the first
    time. That way we aren't sending unnecessary api requests
    when the user doesn't need us to.
    """

    @property
    def data(cls):
        if cls._user_data != None:
            return cls._user_data

        if cls._data == None:
            try:
                cls._data = dict(requests.get(cls._url).json())
            except Exception as e:
                print("Can't find crime library sources.\nEither you don't have internet, or Github's API isn't working.")
                return
        return cls._data
    

    @property
    def url(cls):
        return cls._url



class Library(object, metaclass=MetaLibrary):
    """
    A simple class to interact with the datasources defined
    in a json file stored in the cloud.
    """

    _url = "https://raw.githubusercontent.com/ryayoung/crime/main/colorado-crime-datasets-doc.json"
    _data = None # default sources
    _user_data = None # user-defined sources
    _cache = dict()
    _meta_cache = dict()

    # Try to load default sources online
    try:
        _data = dict(requests.get(_url).json())
    except Exception as e:
        _data = None
    

    @classmethod
    def set_data(cls, data:dict):
        if type(data) != dict:
            raise ValueError("""
Must provide a dict of dicts. Example:
{
    "my_dataset_1": {
        "id": "ab3c-e4gh",
        "base_url": "data.colorado.gov"
    },
    "my_dataset_2": {
        "id": "..."
        "base_url": "..."
    },
    etc...
}""")

        for k, v in data.items():
            if type(v) != dict:
                raise ValueError(f"The value of '{k}' must be a dict with at least two items, 'id' and 'base_url'")
            if "id" not in v:
                raise ValueError(f"Item '{k}' must contain an 'id' element.")
            if "base_url" not in v:
                raise ValueError(f"Item '{k}' must contain a 'base_url' element.")
            
        cls._user_data = data
    

    @classmethod
    def reset_data(cls):
        cls._user_data = None


    @classmethod
    def tabular(cls) -> pd.DataFrame:
        """
        Gives the end user the contents of the self.data dictionary
        as a pandas dataframe, including only the important fields,
        and excluding base_url and id.
        """
        return pd.DataFrame([
            {
                'Name': k,
                'Topic': v.get('topic', np.NaN),
                'Location': v.get('location', np.NaN),
                'Rows': v.get('rows', np.NaN),
                'Type': v.get('type', np.NaN),
                'From': v.get('date_range', [np.NaN, np.NaN])[0],
                'To': v.get('date_range', [np.NaN, np.NaN])[1],
                'Full Name': v.get('full_name', np.NaN),
                'URL': v.get('web_url', np.NaN),
            } for k, v in cls.data.items()
        ]).set_index('Name')
    

    @classmethod
    def cache_add(cls, name, df) -> bool:
        """
        Adds a fully loaded dataframe to cache
        """

        if name not in cls._cache:
            cls._cache[name] = df
            return True

        return False
    

    @classmethod
    def cache_get(cls, name) -> pd.DataFrame:
        """
        Returns dataframe stored in cache
        """
        return cls._cache.get(name, pd.DataFrame()).copy()


    @classmethod
    def meta_cache_add(cls, name, meta) -> bool:
        """
        Adds metadata for particular dataset to cache
        """

        if name not in cls._meta_cache:
            cls._meta_cache[name] = meta
            return True

        return False
    

    @classmethod
    def meta_cache_get(cls, name) -> dict:
        """
        Returns metadata stored in cache
        """
        return cls._meta_cache.get(name, None)



# Maintainer:     Ryan Young
# Last Modified:  May 08, 2022
import requests
import pandas as pd
import numpy as np
import textwrap as tw
from crime.library import Library
from crime.soda_api import Soda

# Declare your app token (optional, recommended)
# >>> cr.set_token("XXXXXXX")

def help():
    print("""DataFrame with info on all datasets:
>>> cr.sources()

Details on a dataset, and description of all columns
>>> cr.sources('dataset_name')

Quickly preview its first 5 rows
>>> cr.load('dataset_name')

Load full dataset
>>> cr.load('dataset_name', full=True)

Declare your own collection of sources (pass a dictionary)
>>> cr.set_sources({
        'district_arrests': { # this is the nickname you'll refer to
            "id": "2e5i-5hfy",
            "base_url": "data.colorado.gov"
        },
        'district_crime': {
            "id": "ya69-n6ta",
            "base_url": "data.colorado.gov"
        },
        # etc...
    }
)

Revert to using the default sources
>>> cr.reset_sources()
""")


def set_token(token):
    """
    Sets the app token to be used for Socrata API
    """
    Soda.token = token


def isnum(val):
    """
    The only fool-proof way to test if val is a number
    and keep it an integer if int, or rounded float if decimal
    """
    val = str(val)
    try:
        return int(val)
    except Exception:
        try:
            return round(float(val), 2)
        except Exception:
            return None


def metadata(name:str) -> dict:
    """
    Returns re-formatted metadata on a source
    as a dictionary. The API metadata is modified to
    include the contents of Library.data, and NOT include
    all the unneccessary bullshit the API gives us.
    """

    cached = Library.meta_cache_get(name)
    if cached != None:
        return cached

    if name not in Library.data:
        return dict()
    
    if Soda.token == None:
        print("Use 'cr.set_token(my_app_token)' to get full api access and avoid this warning")

    # Basic info already stored in Library, but not in metadata
    info = Library.data[name]
    out = dict()

    # Not using dictionary comprehension here, since we're handling
    # different variables differently
    out['id'] = info['id']
    out['base_url'] = info['base_url']
    out['from_year'] = info.get('date_range', [None, None])[0]
    out['to_year'] = info.get('date_range', [None, None])[1]
    out['full_name'] = info.get('full_name', None)
    out['rows'] = info.get('rows', None)
    out['web_url'] = info.get('web_url', None)
    out['location'] = info.get('location', None)
    out['type'] = info.get('type', None)
    out['topic'] = info.get('topic', None)
    out['num_columns'] = None
    out['description'] = None
    out['columns'] = []

    # Socrata metadata
    meta = Soda.get_metadata(info['base_url'], info['id'])

    out['name'] = meta['name']
    out['num_columns'] = len(meta['columns'])
    out['description'] = meta['description']

    for c in meta['columns']:
        cached = c.get('cachedContents', {})
        outc = dict()

        # Occasionally the api hasn't processed display names for
        # for fields, so we have to do it ourself: Split name into
        # separate words by underscore and capitalize each word
        outc['name'] = c['name']
        if c['name'].islower():
            outc['name'] = " ".join([w.title() for w in c['name'].split('_')])

        outc['field'] = c['fieldName']
        outc['type'] = c['dataTypeName']
        outc['null'] = isnum(cached.get('null', None))
        outc['non_null'] = isnum(cached.get('non_null', None))

        outc['avg'] = isnum(cached.get('average', None))
        outc['max'] = isnum(cached.get('largest', None))
        outc['min'] = isnum(cached.get('smallest', None))
        outc['sum'] = isnum(cached.get('sum', None))

        outc['items'] = []

        # If it's a text column, print top items and their frequencies
        if c['dataTypeName'] == 'text':
            if len(cached.get('top', [])) > 1:
                # Check if the text variable is ACTUALLY categorical (IDs, for example, are NOT)
                # To validate, make sure the first couple characters aren't numbers
                if not str(cached.get('top')[1]['item'])[0:1].isdigit():
                    outc['items'] = [
                            {'item': i['item'], 'count': i.get('count', None)}
                        for i in c.get('cachedContents', {}).get('top', [])
                    ]

        out['columns'].append(outc)

    Library.meta_cache_add(name, out)
    return out


def sources(name:str = None) -> pd.DataFrame or None:
    """
    Returns info on available sources.
    ---
    If no arguments are given:
    - Returns pd.DataFrame with all the important details
      already known, without making any requests to Socrata
    ---
    If name is given:
    - Prints a full description of the source and each of its
      columns, including aggregate statistics for numeric columns
      and a list of categories for categorical columns.
    - Returns None
    """

    if not name:
        return Library.tabular()

    else:
        meta = metadata(name)
        if meta == dict():
            return

        print(meta['name'])
        if meta['web_url']:
            print(meta['web_url'], "\n")
        # Wrap description text at 70 characters. (tw.wrap returns a list)
        print(*tw.wrap(meta['description'], width=70), sep="\n")
        print()
        # These likely won't be present if the user defined their own sources
        if meta['rows']:
            print(f"Rows:    {meta['rows']}")
        if meta['num_columns']:
            print(f"Cols:    {meta['num_columns']}")
        if meta['from_year'] and meta['to_year']:
            print(f"Period:  {meta['from_year']} to {meta['to_year']}")
        print()
        print("COLUMNS:")
        print("-------")
        for c in meta['columns']:
            cached = c.get('cachedContents', {})
            print(c['name'])
            print(f"  Field:  {c['field']}")
            print(f"  Type:   {c['type']}")
            null = f"{c['null']:,}" if c['null'] else '-'
            non_null = f"{c['non_null']:,}" if c['non_null'] else '-'
            print(f"  Null:   {null}")
            print(f"  Count:  {non_null}")

            # If it's numeric, print stats.
            if c['type'] == 'number':
                for num in ['min', 'max', 'avg', 'sum']:
                    val = f"{c[num]:,}" if c[num] else '-'
                    print(f"  {num.title()}:    {val}")

            # If it's a text column, print top items and their frequencies
            elif c['type'] == 'text':
                for i in c['items']:
                    count = f"{int(i['count']):,}" if i['count'] else '-'
                    print(f"     {i['item']}  ({count})")

            print()

        return


def columns(name:str) -> pd.DataFrame:
    """
    Returns dataframe with stats on each column.
    Includes all column data from metadata
    EXCEPT the 'items' array for categorical columns.
    """
    col_data = metadata(name)['columns']
    for c in col_data:
        if 'items' in c:
            c.pop('items')

    df = pd.DataFrame(col_data).set_index('field')
    df.index.name = None
    df.columns = df.columns.str.title()
    return df


def load(name:str, **kwargs) -> pd.DataFrame:
    """
    Loads a dataset, and accepts any additional keyword
    arguments for sodapy's get() function.
    ---
    Use the 'limit = n' argument to limit the number
    of rows returned. Useful for large datasets.
    ---
    Returns: pd.DataFrame
    """
    if name not in Library.data:
        return
    
    df = Library.cache_get(name)
    if not df.empty:
        return df

    if Soda.token == None:
        print("Use 'cr.set_token(my_app_token)' to get full api access and avoid this warning")

    info = Library.data[name]
    base_url = info['base_url']
    data_id = info['id']

    try:
        df = Soda.get(base_url, data_id, **kwargs)
        if kwargs.get('full', False) == True:
            Library.cache_add(name, df)
        return df
    except Exception as e:
        print(e)



def df(name:str) -> pd.DataFrame:
    return Library.cache_get(name)



def set_sources(data:dict):
    Library.set_data(data)

def reset_sources():
    Library.reset_data()

def get_current_token():
    return Soda.token








































































































































    
def set_top_secret_token():
    print("No need to set token anymore! A default one is already defined.")
    Soda.token = "POb9E2hovuHBBgdVFLt18TtEG"



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name='crime',
    version='0.1.0',
    description="Explore, load, and get documentation for Colorado crime data.",
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/ryayoung/crime',
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
    author="Ryan Young",
    author_email='ryanyoung99@live.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    keywords='opendata socrata colorado crime',
    install_requires=[
          'pandas',
          'requests',
          'sodapy',
    ],
    python_requires='>=3.9'


)



import requests
from crime.crime import (
    Library,
    help,
    set_token,
    metadata,
    sources,
    columns,
    load,
    df,
    set_sources,
    reset_sources,
    get_current_token,
    set_top_secret_token,
)
from crime.soda_api import Soda

"""
Since we're declaring app token for the user, we should have a backup token in case
the main one fails.
---
Additionally, both of these tokens should be declared in an external file, in case they
need to be changed quickly.
"""
try:
    tokens = requests.get("https://raw.githubusercontent.com/ryayoung/crime/main/cloud_variables.json").json()
    Soda.token = tokens['app_token']
    Soda.alt_token = tokens.get('app_token_alt', None)
except Exception as e:
    Soda.token = None
    Soda.alt_token = None
